/*

This file is part of the Fuzion language implementation.

The Fuzion language implementation is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License as published
by the Free Software Foundation, version 3 of the License.

The Fuzion language implementation is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
License for more details.

You should have received a copy of the GNU General Public License along with The
Fuzion language implementation.  If not, see <https://www.gnu.org/licenses/>.

*/

/*-----------------------------------------------------------------------
 *
 * Tokiwa Software GmbH, Germany
 *
 * Source of class Type
 *
 *---------------------------------------------------------------------*/

package dev.flang.ast;

import dev.flang.util.Errors;
import dev.flang.util.FuzionConstants;
import dev.flang.util.HasSourcePosition;
import dev.flang.util.List;
import dev.flang.util.SourcePosition;


/**
 * Type represents the abstract syntax tree of a Fuzion type parsed from source
 * code.
 *
 * @author Fridtjof Siebert (siebert@tokiwa.software)
 */
public class Type extends AbstractType
{

  //  static int counter;  {counter++; if ((counter&(counter-1))==0) { System.out.println("######################"+counter+" "+this.getClass()); if(false)Thread.dumpStack(); } }

  /*----------------------------  constants  ----------------------------*/


  /**
   * Pre-allocated empty type list. NOTE: There is a specific empty type List
   * Call.NO_GENERICS which is used to distinguish "a.b<>()" (using Type.NONE)
   * from "a.b()" (using Call.NO_GENERICS).
   */
  public static final List<AbstractType> NONE = new List<AbstractType>();
  static { NONE.freeze(); }


  /**
   * pre-allocated empty array of types
   */
  static final Type[] NO_TYPES = new Type[0];

  /**
   * Is this type explicitly a reference or a value type, or whatever the
   * underlying feature is?
   */
  public enum RefOrVal
  {
    Boxed,                  // this is boxed value type or an explicit reference type
    Value,                  // this is an explicit value type
    LikeUnderlyingFeature,  // this is ref or value as declared for the underlying feature
    ThisType,               // this is the type of featureOfType().this.type, i.e., it may be an heir type
  }


  /*----------------------------  variables  ----------------------------*/


  /**
   * The sourcecode position of this type, used for error messages.
   */
  public final HasSourcePosition pos;
  public SourcePosition pos2BeRemoved() { return pos.pos(); }


  /**
   * Is this an explicit reference or value type?  Ref/Value to make this a
   * reference/value type independent of the type of the underlying feature
   * defining a ref type or not, false to keep the underlying feature's
   * ref/value status.
   */
  RefOrVal _refOrVal;


  /**
   * the name of this type.  For a type 'map<string,i32>.entry', this is just
   * the base name 'entry'. For a type parameter 'A', this is 'A'. For an
   * artificial type, this is one of Types.INTERNAL_NAMES (e.g., '--ADDRESS--).
   */
  public String name;
  public String name()
  {
    return name;
  }


  /**
   *
   */
  List<AbstractType> _generics;
  public final List<AbstractType> generics() { return _generics; }


  /**
   * The outer type, for the type p.q.r in the code
   *
   * a
   * {
   *   b
   *   {
   *     c
   *     {
   *       x p.q.r;
   *     }
   *   }
   *   p { ... }
   * }
   *
   * the _outer of "r" is "p.q", and the outer of "q" is "p".
   *
   * However, if p is declared in a, after type resolution, the outer type of
   * "p" is "a" or maybe an heir of "a".
   */
  private AbstractType _outer;


  /**
   * Cached result of outer(). Note the difference to _outer: _outer is the
   * outer type shown in the source code, while outer()/outerCache_ is the
   * actual outer type taken from the type of the outer feature of this type's
   * feature.
   */
  AbstractType _outerCache;


  /**
   *
   */
  AbstractFeature feature;


  /**
   * In case this is the name of a generic argument in an outer feature, this
   * will be set to that generic argument during findGenerics.  Knowing the
   * generics is a pre-requisite to resolving types.
   */
  Generic generic;


  /**
   * For debugging only: Make sure the findGenerics was called before
   * replaceGeneric or resolve is called.
   */
  boolean checkedForGeneric = false;
  boolean checkedForGeneric() { return checkedForGeneric; }


  /**
   * Cached result of calling Types.intern(this).
   */
  Type _interned = null;


  /*--------------------------  constructors  ---------------------------*/


  /**
   * Constructor
   *
   * @param n
   *
   * @param g
   *
   * @param o
   */
  public Type(SourcePosition pos, String n, List<AbstractType> g, AbstractType o)
  {
    this(pos, n,g,o,null, RefOrVal.LikeUnderlyingFeature);
    if (_outer != null && _outer.isRef())
      {
        AstErrors.outerTypeMayNotBeRefType(this);
        this.name = Errors.ERROR_STRING;
      }
  }


  /**
   * Constructor to create a type from an existing type after formal generics
   * have been replaced in the generics arguments and in the outer type.
   *
   * @param t the original type
   *
   * @param g the actual generic arguments that replace t.generics
   *
   * @param o the actual outer type, or null, that replaces t.outer
   */
  public Type(Type t, List<AbstractType> g, AbstractType o)
  {
    this(t.pos2BeRemoved(), t.name, g, o, t.feature, t._refOrVal, false);

    if (PRECONDITIONS) require
      (Errors.count() > 0 ||  (t.generics() instanceof FormalGenerics.AsActuals   ) || t.generics().size() == g.size(),
       Errors.count() > 0 || !(t.generics() instanceof FormalGenerics.AsActuals aa) || aa.sizeMatches(g),
        t == Types.t_ERROR || (t.outer() == null) == (o == null));

    checkedForGeneric = t.checkedForGeneric();
  }


  /**
   * Constructor to create a type from an existing type after formal generics
   * have been replaced in the generics arguments and in the outer type.
   *
   * @param t the original type
   *
   * @param g the actual generic arguments that replace t.generics
   *
   * @param o the actual outer type, or null, that replaces t.outer
   */
  public Type(AbstractType t, List<AbstractType> g, AbstractType o)
  {
    this(t, g, o, false);

    if (PRECONDITIONS) require
      ( (t.generics() instanceof FormalGenerics.AsActuals   ) || t.generics().size() == g.size(),
       !(t.generics() instanceof FormalGenerics.AsActuals aa) || aa.sizeMatches(g),
        t == Types.t_ERROR || (t.outer() == null) == (o == null));
  }


  /**
   * Constructor to create a type from an existing type after formal generics
   * have been replaced in the generics arguments and in the outer type.
   *
   * @param t the original type
   *
   * @param g the actual generic arguments that replace t.generics
   *
   * @param o the actual outer type, or null, that replaces t.outer
   *
   * @param fixOuterThisType NYI: CLEANUP: #737, see below, unclear why this is needed.
   */
  public Type(AbstractType t, List<AbstractType> g, AbstractType o, boolean fixOuterThisType)
  {
    this(t.pos2BeRemoved(),
         t.name(),
         g,
         o,
         t instanceof Type tt ? tt.feature   : t.featureOfType(),

         // NYI: CLEANUP: This seems unnecessarily complex:
         t instanceof Type tt ? tt._refOrVal : (t.isRef() == t.featureOfType().isThisRef() ? RefOrVal.LikeUnderlyingFeature :
                                                t.isRef() ? RefOrVal.Boxed
                                                : RefOrVal.Value),

         fixOuterThisType);

    if (PRECONDITIONS) require
      ( (t.generics() instanceof FormalGenerics.AsActuals   ) || t.generics().size() == g.size(),
       !(t.generics() instanceof FormalGenerics.AsActuals aa) || aa.sizeMatches(g),
        t == Types.t_ERROR || (t.outer() == null) == (o == null));

    checkedForGeneric = t.checkedForGeneric();
  }


  /**
   * Constructor to create a type from an existing type after formal generics
   * have been replaced in the generics arguments and in the outer type.
   *
   * @param t the original type
   *
   * @param o the actual outer type, or null, that replaces t.outer
   */
  public static Type newType(AbstractType t, AbstractType o)
  {
    if (PRECONDITIONS) require
      (t == Types.t_ERROR || (t.outer() == null) == (o == null));

    Type result;
    if (t == Types.t_ERROR ||
        o == Types.t_ERROR   )
      {
        result = Types.t_ERROR;
      }
    else
      {
        result = new Type(t.pos2BeRemoved(), t.featureOfType().featureName().baseName(), t.generics(), o, t.featureOfType(),
                          t.isRef() == t.featureOfType().isThisRef() ? RefOrVal.LikeUnderlyingFeature :
                          t.isRef()                                  ? RefOrVal.Boxed
                                                                     : RefOrVal.Value,
                          false);

        if (t.checkedForGeneric())
          {
            result.checkedForGeneric = true;
          }
      }
    return result;
  }


  /**
   * Constructor
   *
   * @param n
   *
   * @param g the actual generic arguments
   *
   * @param o
   *
   * @param f if this type corresponds to a feature, then this is the
   * feature, else null.
   *
   * @param ref true iff this type should be a ref type, otherwise it will be a
   * value type.
   */
  public Type(HasSourcePosition pos, String n, List<AbstractType> g, AbstractType o, AbstractFeature f, RefOrVal refOrVal)
  {
    this(pos, n, g, o, f, refOrVal, true);
  }


  /**
   * Constructor
   *
   * @param n
   *
   * @param g the actual generic arguments
   *
   * @param o
   *
   * @param f if this type corresponds to a feature, then this is the
   * feature, else null.
   *
   * @param ref true iff this type should be a ref type, otherwise it will be a
   * value type.
   */
  public Type(HasSourcePosition pos, String n, List<AbstractType> g, AbstractType o, AbstractFeature f, RefOrVal refOrVal, boolean fixOuterThisType)
  {
    if (PRECONDITIONS) require
      (pos != null,
       n.length() > 0,
       Errors.count() > 0 || f == null || f.generics().sizeMatches(g == null ? NONE : g));

    this.pos = pos;
    this.name  = n;
    this._generics = ((g == null) || g.isEmpty()) ? NONE : g;
    this._generics.freeze();

    if (fixOuterThisType && o instanceof Type ot && ot.isThisType())
      {
        // NYI: CLEANUP: #737: Undo the asThisType() calls done in This.java for
        // outer types. Is it possible to not create asThisType() in This.java
        // in the first place?
        o = new Type(ot, RefOrVal.LikeUnderlyingFeature);
      }
    this._outer = o;
    this.feature = f;
    this.generic = null;
    this._refOrVal = refOrVal;
    this.checkedForGeneric = f != null;
  }


  /**
   * Constructor for a direct use of a generic argument. For a feature
   *
   *   feat<A,B,c>
   *   {
   *     inner { ...outer... }
   *   }
   *
   * the type of the outer reference within inner is feat<A, B, C>.  This
   * constructor is used to create A, B, C in this case.
   *
   * @param g the formal generic this refers to
   */
  public Type(HasSourcePosition pos, Generic g)
  {
    if (PRECONDITIONS) require
      (g != null);

    this.pos = pos;
    this.name  = g.name();
    this._generics = NONE;
    this._outer = null;
    this.feature = null;
    this.generic = g;
    this._refOrVal = RefOrVal.LikeUnderlyingFeature;
    this.checkedForGeneric = true;
  }


  /**
   * Constructor for built-in types
   *
   * @param n the name, such as "int", "bool".
   */
  public Type(String n)
  {
    this(false, n);
  }


  /**
   * Constructor for built-in types
   *
   * @param ref true iff we create a ref type
   *
   * @param n the name, such as "int", "bool".
   */
  private Type(boolean ref, String n)
  {
    if (PRECONDITIONS) require
      (n.length() > 0);

    this.pos               = SourcePosition.builtIn;
    this.name              = n;
    this._generics         = NONE;
    this._outer            = null;
    this.feature           = null;
    this._refOrVal         = ref ? RefOrVal.Boxed
                                 : RefOrVal.LikeUnderlyingFeature;
    this.checkedForGeneric = true;
  }

  /**
   * Constructor for built-in types
   *
   * @param n the name, such as "int", "bool".
   */
  public static AbstractType type(Resolution res, String n, AbstractFeature universe)
  {
    if (PRECONDITIONS) require
      (n.length() > 0);

    return type(res, false, n, universe);
  }

  /**
   * Constructor for built-in types
   *
   * @param ref true iff we create a ref type
   *
   * @param n the name, such as "int", "bool".
   */
  public static AbstractType type(Resolution res, boolean ref, String n, AbstractFeature universe)
  {
    if (PRECONDITIONS) require
      (n.length() > 0);

    return new Type(ref, n).resolve(res, universe);
  }


  /**
   * Create a ref or value type from a given value / ref type.
   *
   * @param original the original value type
   *
   * @param refOrVal must be RefOrVal.Boxed or RefOrVal.Val
   */
  public Type(Type original, RefOrVal refOrVal)
  {
    if (PRECONDITIONS) require
      (refOrVal != original._refOrVal);

    this.pos                = original.pos;
    this._refOrVal          = refOrVal;
    this.name               = original.name;
    this._generics          = original._generics;
    this._outer             = original._outer;
    this.feature            = original.feature;
    this.generic            = original.generic;
    this.checkedForGeneric  = original.checkedForGeneric;
  }


  /**
   * Create a clone of original that uses originalOuterFeature as context to
   * look up features the type is built from.
   *
   * @param original the original value type
   *
   * @param originalOuterFeature the original feature, which is not a type
   * feature.
   */
  private Type(Type original, AbstractFeature originalOuterFeature)
  {
    this.pos                = original.pos;
    this._refOrVal          = original._refOrVal;
    this.name               = original.name;
    if (original._generics.isEmpty())
      {
        this._generics          = original._generics;
      }
    else
      {
        this._generics = new List<>();
        for (var g : original._generics)
          {
            var gc = (g instanceof Type gt)
              ? gt.clone(originalOuterFeature)
              : g;
            this._generics.add(gc);
          }
        this._generics.freeze();
      }
    this._outer             = (original._outer instanceof Type ot) ? ot.clone(originalOuterFeature) : original._outer;
    this.feature            = original.feature;
    this.generic            = original.generic;
    this.checkedForGeneric  = original.checkedForGeneric;
  }


  /*-----------------------------  methods  -----------------------------*/


  /**
   * This method usually just returns currentOuter. Only for clone()d types that
   * are used in a different outer context, this permits to look up features the
   * type is based on in the original context.
   */
  AbstractFeature originalOuterFeature(AbstractFeature currentOuter)
  {
    return currentOuter;
  }


  /**
   * Create a clone of this Type that uses originalOuterFeature as context to
   * look up features the type is built from.  Generics will be looked up in the
   * current context.
   *
   * This is used for type features that use types from the original feature,
   * but needs to replace generics by the type feature's generics.
   *
   * @param originalOuterFeature the original feature, which is not a type
   * feature.
   */
  Type clone(AbstractFeature originalOuterFeature)
  {
    return
      new Type(this, originalOuterFeature)
      {
        AbstractFeature originalOuterFeature(AbstractFeature currentOuter)
        {
          return originalOuterFeature;
        }
      };
  }


  /**
   * For a type that is not a type parameter, create a new variant using given
   * actual generics and outer type.
   *
   * @param g2 the new actual generics to be used
   *
   * @param o2 the new outer type to be used (which may also differ in its
   * actual generics).
   *
   * @return a new type with same featureOfType(), but using g2/o2 as generics
   * and outer type.
   */
  public AbstractType applyTypePars(List<AbstractType> g2, AbstractType o2)
  {
    if (PRECONDITIONS) require
      (!isGenericArgument());

    return new Type(this, g2, o2);
  }


  /**
   * Create a Types.intern()ed reference variant of this type.  Return this
   * in case it is a reference already.
   */
  public AbstractType asRef()
  {
    if (PRECONDITIONS) require
      (this == Types.intern(this));

    AbstractType result = this;
    if (!isRef() && this != Types.t_ERROR)
      {
        result = Types.intern(new Type(this, RefOrVal.Boxed));
      }
    return result;
  }


  /**
   * Create a Types.intern()ed this.type variant of this type.  Return this
   * in case it is a this.type or a choice variant already.
   */
  public AbstractType asThis()
  {
    if (PRECONDITIONS) require
      (this == Types.intern(this),
       !isGenericArgument());

    AbstractType result = this;
    if (!isThisType() && !isChoice() && this != Types.t_ERROR && this != Types.t_ADDRESS)
      {
        result = Types.intern(new Type(this, RefOrVal.ThisType));
      }

    if (POSTCONDITIONS) ensure
      (result == Types.t_ERROR || result == Types.t_ADDRESS || result.isThisType() || result.isChoice(),
       !(isThisType() || isChoice()) || result == this);

    return result;
  }


  /**
   * Create a Types.intern()ed value variant of this type.  Return this
   * in case it is a value already.
   */
  public AbstractType asValue()
  {
    if (PRECONDITIONS) require
      (this == Types.intern(this));

    AbstractType result = this;
    if (isRef() && this != Types.t_ERROR)
      {
        result = Types.intern(new Type(this, RefOrVal.Value));
      }
    return result;
  }


  /**
   * Call Constructor for a function type that returns a result
   *
   * @param returnType the result type.
   *
   * @param arguments the arguments list
   *
   * @return a Type instance that represents this function
   */
  public static Type funType(SourcePosition pos, AbstractType returnType, List<AbstractType> arguments)
  {
    if (PRECONDITIONS) require
      (returnType != null,
       arguments != null);

    // This is called during parsing, so Types.resolved.f_function is not set yet.
    return new Type(pos,
                    arguments.size() == 1 ? Types.UNARY_NAME : Types.FUNCTION_NAME,
                    new List<AbstractType>(returnType, arguments),
                    null);
  }


  /**
   * setRef is called by the parser when parsing a type of the form "ref
   * <simpletype>".
   */
  public void setRef()
  {
    if (PRECONDITIONS) require
      (this._refOrVal == RefOrVal.LikeUnderlyingFeature);

    this._refOrVal = RefOrVal.Boxed;
  }


  /**
   * isRef
   */
  public boolean isRef()
  {
    return switch (this._refOrVal)
      {
      case Boxed                -> true;
      case Value                -> false;
      case LikeUnderlyingFeature-> ((feature != null) && feature.isThisRef());
      case ThisType             -> false;
      };
  }


  /**
   * isThisType
   */
  public boolean isThisType()
  {
    return this._refOrVal == RefOrVal.ThisType;
  }


  /**
   * setOuter
   *
   * @param t
   */
  void setOuter(Type t)
  {
    if (this._outer == null)
      {
        if (CHECKS) check
          (_interned == null);

        this._outer = t;
      }
    else
      {
        this._outer.setOuter(t);
      }
  }


  /**
   * Get a String representation of this Type.
   *
   * Note that this does not work for instances of Type before they were
   * resolved.  Use toString() for creating strings early in the front end
   * phase.
   */
  public String asString()
  {
    if (PRECONDITIONS) require
      (checkedForGeneric());

    return Types.INTERNAL_NAMES.contains(name)
      ? toString()         // internal types like Types.t_UNDEFINED, t_ERROR, t_ADDRESS
      : super.asString();
  }


  /**
   * toString
   *
   * @return
   */
  public String toString()
  {
    String result;

    if (this == Types.t_ERROR)
      {
        result = Errors.ERROR_STRING;
      }
    else if (Types.INTERNAL_NAMES.contains(name))
      {
        result = name;
      }
    else if (generic != null)
      {
        var f = generic.feature();
        var n = name;
        if (f != null)
          {
            var qn = f.qualifiedName();
            if (generic.isThisTypeInTypeFeature())
              {
                qn = qn.substring(0, qn.lastIndexOf(".type"));
                n = "this.type (in type feature)";
              }
            n = qn + "." + n;
          }
        result = n + (this.isRef() ? " (boxed)" : "");
      }
    else if (_outer != null)
      {
        String outer = _outer.toStringWrapped();
        result = ""
          + (outer == "" ||
             outer.equals(FuzionConstants.UNIVERSE_NAME) ? ""
                                                         : outer + ".")
          + (_refOrVal == RefOrVal.Boxed && (feature == null || !feature.isThisRef()) ? "ref " :
             _refOrVal == RefOrVal.Value &&  feature != null &&  feature.isThisRef()  ? "value "
                                                                                      : ""       )
          + (feature == null ? name
             : feature.featureName().baseName());
      }
    else if (feature == null  || feature == Types.f_ERROR)
      {
        result =
          (_refOrVal == RefOrVal.Boxed ? "ref "
                                       : ""       )
          + name;
      }
    else
      {
        result =
          (_refOrVal == RefOrVal.Boxed && (feature == null || !feature.isThisRef()) ? "ref " :
           _refOrVal == RefOrVal.Value &&  feature != null &&  feature.isThisRef()  ? "value "
                                                                                    : ""       )
          + feature.qualifiedName();
      }
    if (isThisType())
      {
        result = result + ".this.type";
      }
    if (_generics != NONE)
      {
        result = result + _generics
          .toString(" ", " ", "", (g) -> g.toStringWrapped());
      }
    return result;
  }


  /**
   * visit all the features, expressions, statements within this feature.
   *
   * @param v the visitor instance that defines an action to be performed on
   * visited objects.
   *
   * @param outerfeat the feature surrounding this expression.
   */
  public AbstractType visit(FeatureVisitor v, AbstractFeature outerfeat)
  {
    return v
      .actionBefore(this, outerfeat)
      .visit2(v, outerfeat);
  }


  /**
   * Helper method for visit() that is used on the result of v.actionBefore() to
   * visit all the features, expressions, statements within this feature.
   *
   * @param v the visitor instance that defines an action to be performed on
   * visited objects.
   *
   * @param outerfeat the feature surrounding this expression.
   */
  private AbstractType visit2(FeatureVisitor v, AbstractFeature outerfeat)
  {
    if ((feature == null) && (generic == null))
      {
        if (_outer != null)
          {
            var t = _outer.visit(v, outerfeat);
            if (t != _outer)
              {
                if (CHECKS) check
                  (_interned == null);

                _outer = _outer.visit(v, outerfeat);
              }
          }
      }
    if (!(_generics instanceof FormalGenerics.AsActuals))
      {
        generics()
          .stream()
          .filter(g -> g instanceof Type)
          .forEach(g -> g.visit(v, outerfeat));
      }
    return v.action(this, outerfeat);
  }


  /**
   * Find all the types used in this that refer to formal generic arguments of
   * this or any of this' outer classes.
   *
   * @param feat the root feature that contains this type.
   */
  void findGenerics(AbstractFeature outerfeat)
  {
    // NYI   if (PRECONDITIONS) require
    //      (!outerfeat.state().atLeast(Feature.State.RESOLVED_DECLARATIONS));

    if ((feature == null) && (generic == null))
      {
        if (outer() != null)
          {
            if (outer().isGenericArgument())
              {
                AstErrors.formalGenericAsOuterType(pos2BeRemoved(), this);
              }
          }
        else
          {
            var o = outerfeat;
            do
              {
                generic = o.getGeneric(name);
                o = o.outer();
              }
            while (generic == null && o != null);

            if ((generic != null) && !_generics.isEmpty())
              {
                AstErrors.formalGenericWithGenericArgs(pos2BeRemoved(), this, generic);
              }
          }
      }

    checkedForGeneric = true;

    if (POSTCONDITIONS) ensure
      (checkedForGeneric);
  }


  /**
   * resolve artificial types t_ERROR, etc.
   */
  public void resolveArtificialType(AbstractFeature feat)
  {
    if (PRECONDITIONS) require
      (feature == null,
       Types.INTERNAL_NAMES.contains(name));

    feature = feat;

    var interned = Types.intern(this);

    if (CHECKS) check
      (interned == this);
  }


  /**
   * resolve 'abc.this.type' within a type feature. If this designates a
   * 'this.type' withing a type feature, then return the type parameter of the
   * corresponding outer type.
   *
   * Example: if this is
   *
   *   b.this.type
   *
   * within a type feature
   *
   *   a.type.b.type.c.d
   *
   * then we replace 'b.this.type' by the type parameter of a.b.type.
   *
   * @param outerfeat the outer feature that this type is declared in.
   */
  Type resolveThisType(AbstractFeature outerfeat)
  {
    if (PRECONDITIONS) require
      (outerfeat != null,
       outerfeat != null && outerfeat.state().atLeast(Feature.State.RESOLVED_DECLARATIONS),
       checkedForGeneric);

    Type result = this;
    var o = outerfeat;
    while (isThisType() && o != null)
      {
        if (isMatchingTypeFeature(o))
          {
            result = new Type(pos2BeRemoved(), new Generic(o.typeArguments().get(0)));
            o = null;
          }
        else
          {
            o = o.outer();
          }
      }
    return result;
  }


  /**
   * Recursive helper for resolveThisType to check if outerfeat is a type
   * feature with the same name as this.
   *
   * @param outerfeat the outer feature that should be compared to this.
   */
  private boolean isMatchingTypeFeature(AbstractFeature outerfeat)
  {
    return outerfeat.isTypeFeature() &&
      (name + "." + FuzionConstants.TYPE_NAME).equals(outerfeat.featureName().baseName()) &&
      (_outer == null                                   ||
       (_outer instanceof Type ot                   &&
        !ot.isThisType()                            &&
        ot.isMatchingTypeFeature(outerfeat.outer())   )    );
  }


  /**
   * resolve this type
   *
   * @param res this is called during type resolution, res gives the resolution
   * instance.
   *
   * @param feat the outer feature that this type is declared in, used
   * for resolution of generic parameters etc.
   */
  AbstractType resolve(Resolution res, AbstractFeature outerfeat)
  {
    if (PRECONDITIONS) require
      (outerfeat != null,
       outerfeat != null && outerfeat.state().atLeast(Feature.State.RESOLVED_DECLARATIONS),
       checkedForGeneric);

    if (!(outerfeat instanceof Feature of && of.isLastArgType(this)))
      {
        ensureNotOpen();
      }
    AbstractType result = this;
    if (!checkedForGeneric)
      {
        findGenerics(outerfeat);
      }
    if (!isGenericArgument())
      {
        resolveFeature(res, outerfeat);
        if (feature == Types.f_ERROR)
          {
            result = Types.t_ERROR;
          }
        else
          {
            if (isThisType() && _generics.isEmpty())
              {
                this._generics = feature.generics().asActuals();
              }
            _generics = FormalGenerics.resolve(res, _generics, outerfeat);
            if (!feature.generics().errorIfSizeOrTypeDoesNotMatch(_generics,
                                                                  this.pos2BeRemoved(),
                                                                  "type",
                                                                  "Type: " + toString() + "\n"))
              {
                result = Types.t_ERROR;
              }
          }
      }
    return Types.intern(result);
  }


  /**
   * For a Type that is not a generic argument, resolve the feature of that
   * type.  Unlike Type.resolve(), this does not check the generic arguments, so
   * this can be used for type inferencing for the actual generics as in a match
   * case.
   *
   * @param feat the outer feature that this type is declared in, used
   * for resolution of generic parameters etc.
   */
  void resolveFeature(Resolution res, AbstractFeature outerfeat)
  {
    if (PRECONDITIONS) require
      (outerfeat != null,
       outerfeat != null && outerfeat.state().atLeast(Feature.State.RESOLVED_DECLARATIONS));

    if (!checkedForGeneric)
      {
        findGenerics(outerfeat);
      }
    if (!isGenericArgument())
      {
        var of = originalOuterFeature(outerfeat);
        if (_outer != null)
          {
            _outer = _outer.resolve(res, of);
            var ot = _outer.isGenericArgument() ?_outer.genericArgument().constraint() : _outer;
            of = ot.featureOfType();
          }
        if (feature == null)
          {
            var fo = res._module.lookupType(pos2BeRemoved(), of, name, _outer == null);
            feature = fo._feature;
            if (_outer == null && !fo._outer.isUniverse())
              {
                _outer = fo._outer.thisType(fo.isNextInnerFixed());
              }
          }
      }
    if (POSTCONDITIONS) ensure
      (isGenericArgument() || feature != null);
  }


  /**
   * isGenericArgument
   *
   * @return
   */
  public boolean isGenericArgument()
  {
    if (PRECONDITIONS) require
      (checkedForGeneric);

    return generic != null;
  }


  /**
   * featureOfType
   *
   * @return
   */
  public AbstractFeature featureOfType()
  {
    if (PRECONDITIONS) require
      (Errors.count() > 0 || !isGenericArgument(),
       Errors.count() > 0 || feature != null);

    return feature != null
      ? feature
      : Types.f_ERROR;
  }


  /**
   * Is this the type of a type feature, e.g., the type of `(list
   * i32).type`. Will return false for an instance of Type for which this is
   * still unknown since Type.resolve() was not called yet.
   *
   * This is redefined here since `feature` might still be null while this type
   * was not resolved yet.
   */
  boolean isTypeType()
  {
    return feature != null && feature.isTypeFeature();
  }


  /**
   * genericArgument gives the Generic instance of a type defined by a generic
   * argument.
   *
   * @return the Generic instance, never null.
   */
  public Generic genericArgument()
  {
    if (PRECONDITIONS) require
      (isGenericArgument());

    Generic result = generic;

    if (POSTCONDITIONS) ensure
      (result != null);

    return result;
  }


  /**
   * outer type, after type resolution. This provides the whole chain of types
   * until Types.resolved.universe.selfType(), while the _outer field ends with
   * the outermost type explicitly written in the source code.
   */
  public AbstractType outer()
  {
    var result = _outerCache;
    if (result == null)
      {
        result = _outer;
        if (result == null)
          {
            if (feature != null && feature.state().atLeast(Feature.State.LOADED))
              {
                var of = feature.outer();
                if (of == null)
                  {
                    return null;
                  }
                else
                  {
                    result = of.selfType();
                  }
              }
            else if (generic != null)
              {
                if (CHECKS) check
                  (Errors.count() > 0);
              }
            if (result != null)
              {
                result = Types.intern(result);
                _outerCache = result;
              }
          }
      }
    return result;
  }


  /**
   * Check if this or any of its generic arguments is Types.t_ERROR.
   */
  public boolean containsError()
  {
    boolean result = false;
    if (this == Types.t_ERROR)
      {
        result = true;
      }
    else
      {
        for (var t: _generics)
          {
            if (CHECKS) check
              (Errors.count() > 0 || t != null);
            result = result || t == null || t.containsError();
          }
      }

    if (POSTCONDITIONS) ensure
      (!result || Errors.count() > 0);

    return result;
  }


  /**
   * Check if this or any of its generic arguments is Types.t_UNDEFINED.
   *
   * @param exceptFirstGenericArg if true, the first generic argument may be
   * Types.t_UNDEFINED.  This is used in a lambda 'x -> f x' of type
   * 'Function<R,X>' when 'R' is unknown and to be inferred.
   */
  public boolean containsUndefined(boolean exceptFirst)
  {
    boolean result = false;
    if (this == Types.t_UNDEFINED)
      {
        result = true;
      }
    else
      {
        for (var t: _generics)
          {
            if (CHECKS) check
              (Errors.count() > 0 || t != null);
            result = result || !exceptFirst && t != null && t.containsUndefined(false);
            exceptFirst = false;
          }
      }

    return result;
  }

}

/* end of file */
